//____________________________________________________________________________
/*!

\class   genie::flux::GSeaAntaresFileFlux

\brief   A flux driver to read incident neutrinos from evt files

\author  Carla Distefano <distefano_c@lns.infn.it>
         LNS-INFN, Catania

\created March 18, 2019

\cpright  Copyright (c) 2019, The KM3NeT Collaboration
          For the full text of the license see $GSEAGEN/LICENSE

*/
//____________________________________________________________________________

#ifdef _ANTARES_ENABLED__


#include <fstream>
#include <cassert>
#include <cstdio>
#include <iostream>
#include <fstream>

#include <TH2D.h>
#include <TH3D.h>

#include <TF1.h>
#include <TF3.h>
#include <TFile.h>

#ifdef _GENIEVERSIONGTEQ3__
#include "Framework/Conventions/Constants.h"
#include "Framework/Messenger/Messenger.h"
#include "Framework/ParticleData/PDGLibrary.h"
#else
#include "Conventions/Constants.h"
#include "Messenger/Messenger.h"
#include "PDG/PDGLibrary.h"
#endif

#include "SeaNuDrivers/GSeaAntaresFileFlux.h"


using namespace std;
using namespace genie;
using namespace genie::flux;
using namespace genie::constants;




//___________________________________________________________________________
GSeaAntaresFileFlux::GSeaAntaresFileFlux(GenParam * GenPar) :
GSeaFileFlux(GenPar)
{

  fEvt = new event();

  if(fGenPar->FluxFiles.size()>1){
    LOG("GSeaAntaresFileFlux", pFATAL) << "Only one input file is allowed, exiting...";
    exit(1);
  }

  if(fGenPar->FluxFiles[0].FileNames.size()>1){
    LOG("GSeaAntaresFileFlux", pFATAL) << "Only one input file is allowed, exiting...";
    exit(1);
  }

  for(unsigned i=0; i<fGenPar->NuList.size(); i++)fPdgCList->push_back(fGenPar->NuList[i]);

  fInFile.open(fGenPar->FluxFiles[0].FileNames[0]);
  if(!fInFile.is_open()){
    LOG("GSeaAntaresFileFlux", pFATAL) << "Error in opening file "<<fGenPar->FluxFiles[0].FileNames[0]<<", exiting...";
    exit(1);
  }



  this->LoadHeaderFile();






  // Configure GFileFlux options
  // set min/max energy:
  this->SetMinEnergy (fGenPar->EvMin * units::GeV);
  this->SetMaxEnergy (fGenPar->EvMax * units::GeV);

  // set min/max costheta:
  this->SetMinCosTheta (fGenPar->CtMin);
  this->SetMaxCosTheta (fGenPar->CtMax);


  LOG("GSeaAntaresFileFlux", pNOTICE)
       << "Instantiating the Antares file flux driver";


}
//___________________________________________________________________________
GSeaAntaresFileFlux::~GSeaAntaresFileFlux()
{

  fInFile.close();

}
//___________________________________________________________________________
void GSeaAntaresFileFlux::LoadHeaderFile()
{

  bool new_r, new_evt;
  unsigned int nRun, nEvt, Type;

  fEvt->read(fInFile);
  fEvt->info(new_r, new_evt, nRun, nEvt, Type);

  if(new_r){

    unsigned nLine;
    string line;
    int aux;

    nLine=fEvt->ndat("cut_seamuon");
    for(unsigned i=0; i<nLine; i++){
      line = fEvt->next("cut_seamuon");
//      sscanf(line.c_str(),"%lf %lf %lf %lf",&fGenPar->EvMin,&fGenPar->EvMax,&fGenPar->CtMin,&fGenPar->CtMax);

        // dobbiamo controllare che emin e emax sono validi....

    }

    nLine=fEvt->ndat("norma");
    line = fEvt->next("norma");
    sscanf(line.c_str(),"%d %lf ",&aux,&fGenPar->NTot);

  }
  else{
    LOG("GSeaAntaresFileFlux", pFATAL) << "Evt file does not have a header, exiting...";
    exit(1);
  }

  fMinEv = fGenPar->EvMin;
  fMaxEv = fGenPar->EvMax;
  fCosThetaMin = fGenPar->CtMin;
  fCosThetaMax = fGenPar->CtMax;
  fPhiMin = 0.;
  fPhiMax = 2.*kPi;

  this->SetMinEnergy(fGenPar->EvMin);
  this->SetMaxEnergy(fGenPar->EvMax);
  this->ResetNFluxNeutrinos();

  return;
}

//___________________________________________________________________________
vector<GSeaTrack> GSeaAntaresFileFlux::ReadFileEvent()
{

  int id = 0;

  bool new_r, new_evt;
  unsigned int nRun, nEvt, Type;

  unsigned nLine;
  string line;

  double RT=-1.;
  if     (fGenPar->RTOpt.compare("proj")==0) RT = 0.;
  else if(fGenPar->RTOpt.compare("can")==0) {
    double Can2m1 = fGenPar->Can2-fGenPar->Can1;
    RT = 0.5*sqrt(Can2m1*Can2m1 + 4.*(fGenPar->Can3)*(fGenPar->Can3) );
  }
  assert(RT>=0.);

  double Xc=0.;
  double Yc=0.;
  double Zc=-fGenPar->SeaBottomRadius+fGenPar->Can1;  // in m
  double R=fGenPar->SeaBottomRadius+fGenPar->SiteDepth;

  RandomGen * rnd = RandomGen::Instance();

  if(fEvt->read(fInFile)==0){

    fEvt->info(new_r, new_evt, nRun, nEvt, Type);
    EventNumber=nEvt;
    if(!new_r){

      GSeaTrack AuxTrack,AuxMom,AuxGrandMom;
      fDistaMax=0.;

      nLine=fEvt->ndat("track_primary");
      line=fEvt->next("track_primary");
      sscanf(line.c_str(),"%*d %lf %lf %lf %lf %lf %lf %lf %lf %d ",&AuxTrack.V1,&AuxTrack.V2,&AuxTrack.V3,&AuxTrack.D1,&AuxTrack.D2,&AuxTrack.D3,&AuxTrack.E,&AuxTrack.T,&AuxTrack.PdgCode);
      AuxTrack.Id = id++;
      AuxTrack.MotherId=-1;
      AuxTrack.Status=200;
      AuxTrack.V1 = -fabs(AuxTrack.V3/AuxTrack.D3)*AuxTrack.D1;
      AuxTrack.V2 = -fabs(AuxTrack.V3/AuxTrack.D3)*AuxTrack.D2;
      AuxTrack.Length = 0.;
      OtherTracks.push_back(AuxTrack);

      nLine=fEvt->ndat("track_sealepton");
      for(unsigned i=0; i<nLine; i++){
        line = fEvt->next("track_sealepton");
        sscanf(line.c_str(),"%*d %lf %lf %lf %lf %lf %lf %lf %lf %d %lf %*f %d %d",&AuxTrack.V1,&AuxTrack.V2,&AuxTrack.V3,&AuxTrack.D1,&AuxTrack.D2,&AuxTrack.D3,&AuxTrack.E,&AuxTrack.T,&AuxTrack.PdgCode,&AuxTrack.Length,&AuxMom.PdgCode,&AuxGrandMom.PdgCode);
        if ( pdg::IsNeutrino(abs(AuxTrack.PdgCode)) && (AuxTrack.E<fGenPar->EvMin || AuxTrack.E>fGenPar->EvMax) ) continue;
        AuxGrandMom.Id       = id++;
        AuxGrandMom.MotherId = OtherTracks[0].Id;
        AuxGrandMom.Status   = -21;
        AuxMom.Id            = id++;
        AuxMom.MotherId      = AuxGrandMom.Id;
        AuxMom.Status        = -21;
        AuxTrack.Id          = id++;
        AuxTrack.MotherId    = AuxMom.Id;
        AuxTrack.Status      = -1;
        OtherTracks.push_back(AuxGrandMom);
        OtherTracks.push_back(AuxMom);
        SecondaryTracks.push_back(AuxTrack);

        // only muons that have reached the observation level (sea) and have sufficient range
        // to reach the can are allowed to affect the drawing area:
        if ( ( pdg::IsMuon(abs(AuxTrack.PdgCode)) ) && ( AuxTrack.Status == -1 ) ) {
          double fDistaCan = fabs((fGenPar->SiteDepth-fGenPar->Can2-AuxTrack.V3)/AuxTrack.D3); // fGenPar->Can2 = Zmax
          double Range = (pow(10.,fGenPar->RangeSW.Eval(log10(AuxTrack.E)))/(fGenPar->RhoSW)) + fGenPar->MuonRangeTolerance;
          if (Range>=fDistaCan) {
            double V2332diff = AuxTrack.V2*OtherTracks[0].V3-AuxTrack.V3*OtherTracks[0].V2;
            double V3113diff = AuxTrack.V3*OtherTracks[0].V1-AuxTrack.V1*OtherTracks[0].V3;
            double V1221diff = AuxTrack.V1*OtherTracks[0].V2-AuxTrack.V2*OtherTracks[0].V1;
            fDistaMax = max(
              fDistaMax,
              sqrt(
                ( V2332diff*V2332diff + V3113diff*V3113diff + V1221diff*V1221diff )
               /
                ( (OtherTracks[0].V1)*(OtherTracks[0].V1) + (OtherTracks[0].V2)*(OtherTracks[0].V2) + (OtherTracks[0].V3)*(OtherTracks[0].V3) )
              )
            );
          }

          LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"fDistaMax:"<<fDistaMax;
          double Can21diff = fGenPar->Can2-fGenPar->Can1;
          LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"fDistaMax limit:"<<Can21diff*Can21diff+4.*(fGenPar->Can3)*(fGenPar->Can3);
        }
      }

      fDistaMax += 100;

      // generate the primary vertex
      double xDif = fX0-Xc;
      double yDif = fY0-Yc;
      double zDif = fZ0-Zc;
      double b  = OtherTracks[0].D1*xDif + OtherTracks[0].D2*yDif + OtherTracks[0].D3*zDif;
      double c  = xDif*xDif+yDif*yDif+zDif*zDif-R*R;
      double RL = fabs(-b-sqrt(b*b-c));

      double x = 0.;
      double y = 0.;
      double z = 0.;

      if (RT==0.) {
        double height = (fGenPar->Can2-fGenPar->Can1+fGenPar->HBedRock) + fDistaMax;
        double radius = fGenPar->Can3 + fDistaMax;
        double AreaTop  = kPi * radius*radius * fabs(OtherTracks[0].D3) ;
        double AreaSide = 2 * height * radius * sqrt( 1. - (OtherTracks[0].D3)*(OtherTracks[0].D3) );
        fAgen = AreaTop + AreaSide;
        if ( rnd->RndFlux().Rndm() < AreaTop / fAgen ) {
          double r   = radius * sqrt( rnd->RndFlux().Rndm() );
          double phi = 2 * kPi * rnd->RndFlux().Rndm();
          x = r*sin(phi);
          y = r*cos(phi);
          z = OtherTracks[0].D3 > 0 ? -height*0.5 : height*0.5;
        }
        else {
          double zaux = height * (rnd->RndFlux().Rndm()-0.5);
          double yaux = 2 * radius * (rnd->RndFlux().Rndm()-0.5);
          double xaux = -sqrt( radius*radius - yaux*yaux );
          TVector3 v(xaux,yaux,zaux);
          double phidir = atan2( OtherTracks[0].D2, OtherTracks[0].D1 );
          v.RotateZ( phidir );
          x = v.X();
          y = v.Y();
          z = v.Z();
        }
        x += -RL * OtherTracks[0].D1;
        y += -RL * OtherTracks[0].D2;
        z += -RL * OtherTracks[0].D3;
      }
      else {
        RT+=fDistaMax;
        fAgen = kPi*RT*RT;
        x += -RL * OtherTracks[0].D1;  // we have to multiply RL times the arrival direction cosines
        y += -RL * OtherTracks[0].D2;
        z += -RL * OtherTracks[0].D3;
        TVector3 vec(x,y,z);                      // vector towards selected point
        TVector3 dvec1 = vec.Orthogonal();        // orthogonal vector
        TVector3 dvec2 = dvec1;                   // second orthogonal vector
        dvec2.Rotate(-kPi*0.50,vec);        // rotate second vector by 90deg, now forming a new orthogonal cartesian coordinate system
        double psi = 2.*kPi* rnd->RndFlux().Rndm(); // rndm angle [0,2pi]
        double random = rnd->RndFlux().Rndm();        // rndm number  [0,1]
        dvec1.SetMag(sqrt(random)*RT*cos(psi));
        dvec2.SetMag(sqrt(random)*RT*sin(psi));
        x += dvec1.X() + dvec2.X();
        y += dvec1.Y() + dvec2.Y();
        z += dvec1.Z() + dvec2.Z();
      }

      x += fX0;
      y += fY0;
      z += fZ0;

      nLine=fEvt->ndat("weights");
      for(unsigned i=0; i<nLine; i++){
        double waux;
        line = fEvt->next("weights");
        sscanf(line.c_str(),"%lf %lf %lf",&waux,&fGenWeight,&fEvtWeight);
        fGenWeight*=fAgen;
        fEvtWeight*=fAgen;

        fEvt->tagd("weights", i);
        char ch[1024];
        sprintf(ch, "%12.3E %12.3E %12.3E",fAgen,fGenWeight,fEvtWeight);
        fEvt->taga("weights",ch);

        sprintf(ch, "%f",fDistaMax);
        fEvt->taga("dista_max",ch);

        sprintf(ch, "%f",RT);
        fEvt->taga("RADD",ch);

      }

      for (auto& Track : OtherTracks) {
        Track.V1 += x;
        Track.V2 += y;
        Track.V3 += z;
        // for muons and neutrinos, we translated the starting point from the generation area up to the sealevel
        if ( pdg::IsNeutrino(abs(Track.PdgCode)) || pdg::IsMuon(abs(Track.PdgCode)) ) {
          double xOff = Track.V1-Xc;
          double yOff = Track.V2-Yc;
          double zOff = Track.V3-Zc;
          double bmu  = Track.D1*xOff+ Track.D2*yOff +Track.D3*zOff;
          double cmu  = xOff*xOff+yOff*yOff+zOff*zOff-R*R;
          double RLmu = -bmu-sqrt(bmu*bmu-cmu);
          Track.V1+=RLmu*Track.D1;
          Track.V2+=RLmu*Track.D2;
          Track.V3+=RLmu*Track.D3;
        }

      }
      
      for (auto& Track : SecondaryTracks) {
        Track.V1 += x;
        Track.V2 += y;
        Track.V3 += z;
        // for muons and neutrinos, we translated the starting point from the generation area up to the sealevel
        if ( pdg::IsNeutrino(abs(Track.PdgCode)) || pdg::IsMuon(abs(Track.PdgCode)) ) {
          double xOff = Track.V1-Xc;
          double yOff = Track.V2-Yc;
          double zOff = Track.V3-Zc;
          double bmu  = Track.D1*xOff+ Track.D2*yOff +Track.D3*zOff;
          double cmu  = xOff*xOff+yOff*yOff+zOff*zOff-R*R;
          double RLmu = -bmu-sqrt(bmu*bmu-cmu);
          Track.V1+=RLmu*Track.D1;
          Track.V2+=RLmu*Track.D2;
          Track.V3+=RLmu*Track.D3;
        }

      }
    }
    
  }
  FullRun = true;     // full run extracted
  return vector<GSeaTrack>(); // we're done
}
//___________________________________________________________________________
void GSeaAntaresFileFlux::ComputeWeights(GSeaEvent * SeaEvt)
{

  SeaEvt->DistaMax=fDistaMax;
  SeaEvt->Agen=fAgen;
  SeaEvt->GenWeight=fGenWeight;
  SeaEvt->EvtWeight=fEvtWeight;

  this->GetEvtTime(SeaEvt->EvtTime,SeaEvt->EvtTime+1);

  return;
}

#endif
