//____________________________________________________________________________
/*!

\class   genie::flux::GSeaCORSIKAFileFlux.cxx

\brief   A flux driver to read incident muons and neutrinos from CORSIKA7 binary files

\author  Piotr Kalaczyński <piotr.kalaczynski@ncbj.gov.pl>
         NCBJ Warsaw

\created 25.12.2020

\cpright  Copyright (c) 2021, The KM3NeT Collaboration
          For the full text of the license see $GSEAGEN/LICENSE

*/
//____________________________________________________________________________

#include <fstream>
#include <cassert>
#include <cstdio>
#include <iostream>
#include <fstream>
#include <unordered_map>
#include <cmath>

#ifdef _GENIEVERSIONGTEQ3__
#include "Framework/Conventions/Constants.h"
#include "Framework/Messenger/Messenger.h"
#include "Framework/ParticleData/PDGLibrary.h"
#else
#include "Conventions/Constants.h"
#include "Messenger/Messenger.h"
#include "PDG/PDGLibrary.h"
#endif

#include "SeaNuDrivers/GSeaCORSIKAFileFlux.h"
#include "utils/array_operations.h"

using namespace std;
using namespace genie;
using namespace genie::flux;
using namespace genie::constants;


//___________________________________________________________________________
GSeaCORSIKAFileFlux::GSeaCORSIKAFileFlux(GenParam * GenPar) :
GSeaFileFlux(GenPar)
{
  if(fGenPar->FluxFiles.size()>1){
    LOG("GSeaCORSIKAFileFlux", pFATAL) << "Only one input file is allowed, exiting...";
    exit(1);
  }

  if(fGenPar->FluxFiles[0].FileNames.size()>1){
    LOG("GSeaCORSIKAFileFlux", pFATAL) << "Only one input file is allowed, exiting...";
    exit(1);
  }

  for(unsigned int i=0; i<fGenPar->NuList.size(); i++)fPdgCList->push_back(fGenPar->NuList[i]);

  LOG("GSeaCORSIKAFileFlux", pINFO) <<"-== Opening CORSIKA binary file: "<< fGenPar->FluxFiles[0].FileNames[0];
  CORSIKAFile.open(fGenPar->FluxFiles[0].FileNames[0]);
  if(!CORSIKAFile.is_open()){
    LOG("GSeaCORSIKAFileFlux", pFATAL) << "Error in opening file "<<fGenPar->FluxFiles[0].FileNames[0]<<", exiting...";
    exit(1);
  }

  CORSIKAFile.read((char*)&StdData,sizeof(StdData)); // (buffer,length)

  // check for the 32/64-bit architecture and standard/thinning format
  if ( 217433.0 < StdData[LenSubblStd+1] && StdData[LenSubblStd+1] < 217433.2 ) {
     LOG("GSeaCORSIKAFileFlux", pNOTICE) <<"-== 32-bit file, standard format ==-";
     LBit = 32;
     LThi = 0;
  }
  else if ( 217433.0 < StdData[LenSubblStd+2] && StdData[LenSubblStd+2] < 217433.2 ) {
     LOG("GSeaCORSIKAFileFlux", pNOTICE) <<"-== 64-bit file, standard format ==-";
     LBit = 64;
     LThi = 0;
  }
  else if ( 217433.0 < StdData[LenSubblThi+1] && StdData[LenSubblThi+1] < 217433.2 ) {
     LOG("GSeaCORSIKAFileFlux", pNOTICE) <<"-== 32-bit file, thinning format ==-";
     LBit = 32;
     LThi = 1;
  }
  else if ( 217433.0 < StdData[LenSubblThi+2] && StdData[LenSubblThi+2] < 217433.2 ) {
     LOG("GSeaCORSIKAFileFlux", pNOTICE) <<"-== 64-bit file, thinning format ==-";
     LBit = 64;
     LThi = 1;
  }
  LBit = (LBit-32) / 32;
  LenSubblock = LenSubblStd + LThi*39;

  CORSIKAFile.clear( );
  CORSIKAFile.seekg( 0, ios::beg ); // go back to the beginning of the file
  iRec=1,lp=1;
  PrimID=0;

  // precompute sin and cos by which we'll rotate:
  sinSiteRot = sin(0.5*kPi + fGenPar->SiteDeclination);
  cosSiteRot = cos(0.5*kPi + fGenPar->SiteDeclination);
  
  fGenPar->SeaBottomRadius = 6371315. - fGenPar->SiteDepth; //  Earth radius as used in CORSIKA [m] - site depth [m]

  LOG("GSeaCORSIKAFileFlux", pNOTICE) << "Instantiating the CORSIKA7 file flux driver";
  LOG("GSeaCORSIKAFileFlux", pNOTICE) << "The SeaBottomRadius value that will be used is "<< fGenPar->SeaBottomRadius <<"m (based on CORSIKA def. and site depth)";

  if (!(fGenPar->SilentCORSIKA)) LOG("GSeaCORSIKAFileFlux", pNOTICE) <<"Reading the header ...";
  FullEvent=false;
  while(!FullEvent){ // iterate until a complete event is extracted
    if (lp >= LenSubblock*NumSubblocks) { // only read if all subblocks are read
      lp=1; // subblock index has to be reset
      // treating the data records according to their format (standard/thinning):
      if ( LThi == 1 ) { // long record (thinning)

         CORSIKAFile.read((char*)&ThiData,sizeof(ThiData));


         if ( LBit == 1 ) {
            CORSIKAFile.read((char*)&ZData,sizeof(ZData));
            ThiData[0] = 0.21345763838967e-41;
            for( unsigned int i=1; i<NumThi; i++ ) ThiData[i] = ThiData[i+LBit];
            ThiData[NumThi] = ZData[0];
         }

         for( unsigned int i=1; i<=NumStd; i++ ) StdData[i] = ThiData[i];
      }
      else if ( LThi == 0 ) { // short record (standard)

         CORSIKAFile.read((char*)&StdData,sizeof(StdData));

         if ( LBit == 1 ) CORSIKAFile.read((char*)&ZData,sizeof(ZData));

         StdData[0] = 0.21345763838967e-41;
         for( unsigned int i=NumStd; i<=NumThi; i++ ) ThiData[i] = 0.;
         for( unsigned int i=1; i<=NumStd; i++ ) ThiData[i] = StdData[i+LBit];
      }
      // to be sure we do not go too far
      if ( (iRec>fGenPar->MaxNRec) && (fGenPar->MaxNRec!=0) ) {
        LOG("GSeaCORSIKAFileFlux", pFATAL) << "Exceeded the maximum number of records, exiting...";
        exit(1);
      }
      iRec++;
    }
    else {

      LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Reading subblock " << ((lp-1)/LenSubblock+1) << " ==-";
      if ( (211285.2 < StdData[lp]) && (StdData[lp] < 211285.3) ) {       // RUNH = 211285.2812500000000000
         // double RunH=StdData[lp]               // normally RUNH = 211285
         nRun = StdData[lp+1];    // run number
         if (nRun == 0){
           LOG("GSeaCORSIKAFileFlux", pFATAL) << "CORSIKA file does not have a header, exiting...";
           exit(1);
         }
         fGenPar->CORSIKADate = StdData[lp+2];    // yymmdd of the run
         fGenPar->CORSIKAVer = StdData[lp+3];     // version of CORSIKA
         // double obslev = StdData[lp+5];          // observation level [cm]
         fGenPar->Alpha = StdData[lp+15];         // spectral index of E_primary
         fGenPar->EPrimMin = StdData[lp+16];      // minimal E_primary
         fGenPar->EPrimMax = StdData[lp+17];      // maximal E_primary
         fGenPar->EvMin = StdData[lp+20];         // minimal E_hadron -> ~ minimal neutrino energy
         // double EeMin = StdData[lp+21];          // minimal E_electron
         fGenPar->EmuMin = StdData[lp+22];        // minimal E_muon
         // double EgammaMin = StdData[lp+23];      // minimal E_gamma
         fGenPar->NTot = StdData[lp+92];          // number of generated showers
      }
      else if ( (217433.0 < StdData[lp]) && (StdData[lp] < 217433.1) ) {  // EVTH = 217433.0781250000000000
         PrimID = StdData[lp+2];                  // particle ID (CORSIKA scheme!)

         fGenPar->Primary=CorsikaToPdg(PrimID);
         fGenPar->PrimCtMin=cos(kPi+StdData[lp+80]*kPi/180.); // max cos(zen) is for the min zen
         fGenPar->PrimCtMax=cos(kPi+StdData[lp+81]*kPi/180.);  // min cos(zen) is for the max zen
         // int numSeeds = StdData[lp+12]; // Number of seeds
         fGenPar->CORSIKASeed[0]=StdData[lp+13]; // CORSIKA seed #1
         fGenPar->CORSIKASeed[1]=StdData[lp+16]; // CORSIKA seed #2
         switch(uint8_t(StdData[lp+74])) { // LE flag
           case 1 : fGenPar->CORSIKAHadronicModelLE = "GHEISHA"; break;
           case 2 : fGenPar->CORSIKAHadronicModelLE = "UrQMD"; break;
           case 3 : fGenPar->CORSIKAHadronicModelLE = "FLUKA"; break;
         }
         switch(uint8_t(StdData[lp+75])) { // HE flag
           case 0 : fGenPar->CORSIKAHadronicModelHE = "HDPM"; break;
           case 1 : fGenPar->CORSIKAHadronicModelHE = "VENUS"; break;
           case 2 : fGenPar->CORSIKAHadronicModelHE = "SIBYLL"; break;
           case 3 : fGenPar->CORSIKAHadronicModelHE = "QGSJET"; break;
           case 4 : fGenPar->CORSIKAHadronicModelHE = "DPMJET"; break;
           case 5 : fGenPar->CORSIKAHadronicModelHE = "NEXUS"; break;
           case 6 : fGenPar->CORSIKAHadronicModelHE = "EPOS"; break;
         }
         switch(uint8_t(StdData[lp+138])) { // SIBYLL interaction version flag
           case 0 : break; // SIBYLL not used
           case 1 : fGenPar->CORSIKAHadronicModelHE.append("-1.6"); break;
           case 2 : fGenPar->CORSIKAHadronicModelHE.append("-2.1"); break;
           case 3 : fGenPar->CORSIKAHadronicModelHE.append("-2.3c"); break;
           case 4 : fGenPar->CORSIKAHadronicModelHE.append("-2.3d"); break;
         }
         switch(uint8_t(StdData[lp+140])) { // QGSJET version flag
           case 0 : break; // QGSJET not used
           case 1 : fGenPar->CORSIKAHadronicModelHE.append("OLD"); break;
           case 2 : fGenPar->CORSIKAHadronicModelHE.append("01d"); break;
           case 3 : fGenPar->CORSIKAHadronicModelHE.append("-II"); break;
         }
         // other models have nothing to be added
      }
      else if ( (3397.3 < StdData[lp]) && (StdData[lp] < 3397.4) ) {      // EVTE =   3397.3918457031250000
         LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Event end (EVTE subblock) found (subblock #" << lp << " record #" << iRec << ") ==-";
         FullEvent = true; // end loop, full event extracted
         lp+=LenSubblock; // move to the next subblock
         break;
      }
      lp+=LenSubblock; // move to the next subblock
    }
  }

  CORSIKAFile.clear( );
  CORSIKAFile.seekg( 0, ios::beg ); // go back to the beginning of the file
  iRec=0,lp=1;
  PrimID=0;
  LOG("GSeaCORSIKAFileFlux", pNOTICE) <<"Finished reading the header";
}
//___________________________________________________________________________
GSeaCORSIKAFileFlux::~GSeaCORSIKAFileFlux()
{

  CORSIKAFile.close();

}
//___________________________________________________________________________
vector<GSeaTrack> GSeaCORSIKAFileFlux::ReadFileEvent()
{

  uint32_t id = 0;
  double RadDep = fGenPar->SeaBottomRadius + fGenPar->SiteDepth;

  GSeaTrack AuxTrack,AuxMother,AuxGrandMother,AuxMuonAddInfo;
  // FullRecord = false;
  bool KeepIterating = true;
  while (KeepIterating){
    FullEvent = false;
    while (!FullEvent){ // iterate until a complete event is extracted
      if ((lp >= LenSubblock*NumSubblocks) || ((lp==1) && (iRec==0)))  { // read if all subblocks of the record are read

        if ( (iRec>fGenPar->MaxNRec) && (fGenPar->MaxNRec!=0) ) { // to be sure we do not go too far
          LOG("GSeaCORSIKAFileFlux", pFATAL) << "Exceeded the maximum number of records, exiting...";
          exit(1);
        }
        LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Reading record " << iRec << " ==-";

        // treating the data records according to their format (standard/thinning):
        if ( LThi == 1 ) { // long record (thinning)

           CORSIKAFile.read((char*)&ThiData,sizeof(ThiData));


           if ( LBit == 1 ) {
              CORSIKAFile.read((char*)&ZData,sizeof(ZData));
              ThiData[0] = 0.21345763838967e-41;
              for( unsigned int i=1; i<NumThi; i++ ) ThiData[i] = ThiData[i+LBit];
              ThiData[NumThi] = ZData[0];
           }

           for( unsigned int i=1; i<=NumStd; i++ ) StdData[i] = ThiData[i];
        }
        else if ( LThi == 0 ) { // short record (standard)

           CORSIKAFile.read((char*)&StdData,sizeof(StdData));

           if ( LBit == 1 ) CORSIKAFile.read((char*)&ZData,sizeof(ZData));

           StdData[0] = 0.21345763838967e-41;
           for( unsigned int i=NumStd; i<=NumThi; i++ ) ThiData[i] = 0.;
           for( unsigned int i=1; i<=NumStd; i++ ) ThiData[i] = StdData[i+LBit];
        }

        lp=1; // subblock index has to be reset
        iRec++;
      }
      else {

        LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Reading subblock " << ((lp-1)/LenSubblock+1) << " ==-";
        if ( (211285.2 < StdData[lp]) && (StdData[lp] < 211285.3) ) {       // RUNH = 211285.2812500000000000
           LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Run header (RUNH subblock) found (subblock #" << lp << " record #" << iRec << ") ==-";

           // double RunH=StdData[lp]               // normally RUNH = 211285
           fGenPar->RunNu = StdData[lp+1];          // run number
           // double corsikaVer = StdData[lp+3];     // version of CORSIKA
           // double obslev = StdData[lp+5];        // observation level [cm]
           fGenPar->Alpha = StdData[lp+15];         // spectral index of E_primary
           fGenPar->EPrimMin = StdData[lp+16];      // minimal E_primary
           fGenPar->EPrimMax = StdData[lp+17];      // maximal E_primary
           fGenPar->EvMin = StdData[lp+20];         // minimal E_hadron -> ~ minimal neutrino energy
           // double EeMin = StdData[lp+21];        // minimal E_electron
           fGenPar->EmuMin = StdData[lp+22];        // minimal E_muon
           // double EgammaMin = StdData[lp+23];    // minimal E_gamma
           fGenPar->NTot = StdData[lp+92];          // number of generated showers
           if (fGenPar->RunNu == 0){
             LOG("GSeaCORSIKAFileFlux", pFATAL) << "CORSIKA file does not have a header, exiting...";
             exit(1);
           }

        }
        else if ( (217433.0 < StdData[lp]) && (StdData[lp] < 217433.1) ) {  // EVTH = 217433.0781250000000000
          if (!(fGenPar->SilentCORSIKA)) LOG("GSeaCORSIKAFileFlux", pINFO) <<"-== Event header (EVTH subblock) found (shower #" << StdData[lp+1] << "/"<<fGenPar->NTot<<" ,record #" << iRec << ") ==-";
          vector<GSeaTrack>().swap(OtherTracks); // clear memory
          vector<GSeaTrack>().swap(SecondaryTracks); // clear memory
          fDistaMax=0.; // resets the DistaMax to be evaluated for the next shower
          
          // Extracting selected information from the file:
          // for more details see CORSIKA Userguide: 10.2 Normal Particle Output
          // (p.124-139)
          // double EvtH=StdData[lp];                    // normally EVTH = 211285
          EventNumber = (unsigned int) StdData[lp+1];  // event (shower) number
          PrimID = StdData[lp+2];                      // particle ID (CORSIKA scheme!)
          EPrim = StdData[lp+3];                       // E_primary [GeV]
          // float StartAlt = StdData[lp+4];             // starting altitude [g/cm^2]
          // unsigned int NumT = StdData[lp+5];          // number of 1st target if fixed
          HInt = StdData[lp+6]*0.01;            // height above obslev (sea level) of 1st interaction [m] (z coord., negative if tracking starts at margin of atmosphere)
          double MomentumPrimX = StdData[lp+7];        // p_x [GeV/c] before interaction
          double MomentumPrimY = StdData[lp+8];        // p_y [GeV/c] before interaction
          double MomentumPrimZ = StdData[lp+9];        // p_z [GeV/c] before interaction (in -z direction!)
          double Zenith = StdData[lp+10];              // zenith [rad]
          // double Azimuth = StdData[lp+11];            // azimuth [rad]
          MinEPrim = StdData[lp+58];                   // min E_primary [GeV]
          MaxEPrim = StdData[lp+59];                   // max E_primary [GeV]

          double TotPrimMomentum = sqrt(MomentumPrimX*MomentumPrimX+MomentumPrimY*MomentumPrimY+MomentumPrimZ*MomentumPrimZ);
          double DirXPrim=MomentumPrimX/TotPrimMomentum;
          double DirYPrim=MomentumPrimY/TotPrimMomentum;
          double DirZPrim=MomentumPrimZ/TotPrimMomentum;
          
          // rotate from CORSIKA to KM3NeT coordinates: (as if the shower started right above the detector)
          
          AuxTrack.D1=cosSiteRot * DirXPrim - sinSiteRot * DirYPrim;  // x direction (already in KM3NeT cooridnates)
          AuxTrack.D2=sinSiteRot * DirXPrim + cosSiteRot * DirYPrim;  // y direction (already in KM3NeT cooridnates)
          AuxTrack.D3=-DirZPrim;                                      // z direction (already in KM3NeT cooridnates)
          
          
          double CZen = cos(Zenith);
          // distance when tracing back up from sea to the 1st interaction:
          AuxTrack.Length =  sqrt(RadDep*RadDep*CZen*CZen+HInt*HInt+2.*fabs(HInt)*RadDep) - RadDep*CZen;
          AuxTrack.V1=-AuxTrack.D1*AuxTrack.Length;                      // x coordinate of 1st interaction [m]
          AuxTrack.V2=-AuxTrack.D2*AuxTrack.Length;
          AuxTrack.V3=-AuxTrack.D3*AuxTrack.Length + fGenPar->SiteDepth;
                                                                      // coordinates of primary are traced back from the sea
                                                                      // and will be later overwritten when shifting showers
                                                                      // (sea position of the primary is always 0,0,0)
          
          AuxTrack.E=EPrim;
          AuxTrack.T=0; // time counter (starts at 1st interaction); always just 0 for primary
          AuxTrack.PdgCode=CorsikaToPdg(PrimID);
          
          AuxTrack.Id = id++;
          AuxTrack.MotherId = -1;
          AuxTrack.Status = 200;      // i.e. primary particle
          
          AuxTrack.GenCounter = 0;
          OtherTracks.push_back(AuxTrack);
          
        }
        else if ( (3397.3 < StdData[lp]) && (StdData[lp] < 3397.4) ) {      // EVTE =   3397.3918457031250000
           LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Event end (EVTE subblock) found ==-";
           
           if (fGenPar->CustomRotation[0]!=0. || fGenPar->CustomRotation[1]!=0. || fGenPar->CustomRotation[2]!=0.) CustomRotation(&OtherTracks[0]);
           
           FullEvent = true;  // end loop, full event extracted
           lp+=LenSubblock;   // move to the next subblock
           break;
        }
        else if ( (3301.3 < StdData[lp]) && (StdData[lp] < 3301.4) ) {      // RUNE =   3301.3325195312500000
          LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Run end (RUNE subblock) found (record #" << iRec << ") ==-";
          FullEvent = true;   // end loop, full event extracted
          FullRun = true;     // full run extracted
          KeepIterating = false;
          lp+=LenSubblock;    // move to the next subblock
          vector<GSeaTrack>().swap(OtherTracks); // clear memory
          vector<GSeaTrack>().swap(SecondaryTracks); // clear memory
          break;
        }
        else if ( (52815.2 < StdData[lp]) && (StdData[lp] < 52815.3) ) {    // LONG =  52815.2968750000000000
          LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Analyzing LONG subblock ... ==-";
        }
        else {                                                              // Particle subblock

          LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Analyzing particle subblock (record #" << iRec <<") ... ==-";
          // const particle_data* p = (const particle_data*) &StdData[lp]; // points the particle_data struct at the beginning of the subblock
          
          int16_t ParticleID=-1;
          // while ( ParticleID!=0 ) { // iterating over particles (up to 4x39)

          for (unsigned int j = 0; j < (MaxNTracksPerSubblock) ; j++) { // iterating over particles (up to 4x39)

            // Extracting selected information from the file:
            // for more details see CORSIKA Userguide: 10.2 Normal Particle Output
            // (p.124-139)
            
            // ParticleID = p->ID/1000;     // pID (CORSIKA scheme!)
            ParticleID = StdData[lp+7*j]/1000;     // pID (CORSIKA scheme!)
            if (ParticleID==0) break;
            
            // int16_t GenCounter = int32_t(p->ID)%1000;   // generation counter
            int16_t GenCounter = int32_t(StdData[lp+7*j])%1000;   // generation counter
            
            double MomentumX = StdData[lp+1+7*j];       // p_x [GeV/c]
            double MomentumY = StdData[lp+2+7*j];       // p_y [GeV/c]
            double MomentumZ = StdData[lp+3+7*j];       // p_z [GeV/c]
            double PosXOrGenCount = StdData[lp+4+7*j];  // x coordinate [cm]
                                                        // or
                                                        // extended electromagnetic generation counter of final particle
            double PosYOrPenDepth = StdData[lp+5+7*j];  // y coordinate [cm]
                                                        // or
                                                        // penetration depth in [g/cm2]
            double PosZOrTime = StdData[lp+6+7*j];      // t [ns] since first interaction
                                                        // or
                                                        // z [cm] of muon production point (for additional muon info (MUADDI))

            // calculate the directions still in CORSIKA coordinate system:
            double TotMomentum = sqrt(MomentumX*MomentumX+MomentumY*MomentumY+MomentumZ*MomentumZ);
            float DirX=MomentumX/TotMomentum;
            float DirY=MomentumY/TotMomentum;
            float DirZ=MomentumZ/TotMomentum;

            // rotate pos and dir from CORSIKA to KM3NeT coordinates:
            float PosX = 0.01*(cosSiteRot * PosXOrGenCount - sinSiteRot * PosYOrPenDepth);  // x position [m]
            float PosY = 0.01*(sinSiteRot * PosXOrGenCount + cosSiteRot * PosYOrPenDepth);  // y position [m]
            // PosZ is handled separately in cases below
            AuxTrack.D1=cosSiteRot * DirX - sinSiteRot * DirY; // x direction
            AuxTrack.D2=sinSiteRot * DirX + cosSiteRot * DirY; // y direction
            AuxTrack.D3=-DirZ; // z direction

            double Mass = CorsikaRestMass(ParticleID);
            AuxTrack.E=sqrt(TotMomentum*TotMomentum+Mass*Mass);

            if (ParticleID>=0) AuxTrack.PdgCode=CorsikaToPdg(ParticleID);
            else AuxTrack.PdgCode=CorsikaToPdg(-ParticleID); // (grand-)mother has negative ID

            switch (ParticleID) {
              // // uncomment for neutrinos to be included for propagation:
              // case 66:                                      // (anti-)neutrino
              // case 67:                                      // (anti-)neutrino
              // case 68:                                      // (anti-)neutrino
              // case 69:                                      // (anti-)neutrino
              // case 133:                                     // (anti-)neutrino
              // case 134:                                     // (anti-)neutrino
              case 5:                                       // (anti-)muon
              case 6:    { AuxTrack.Status = -1; break;  }  // (anti-)muon
              case 85:
              case 86:   { AuxTrack.Status = 22; break; }  // decaying mu at start (muon that does NOT reach the observation level!)
              case 95:
              case 96:   { AuxTrack.Status = 23; break; }  // decaying mu at end   (muon that does NOT reach the observation level!)
              case 71:   { AuxTrack.Status = 24; break; }  // eta -> 2 gamma
              case 72:   { AuxTrack.Status = 25; break; }  // eta -> 3 pi0
              case 73:   { AuxTrack.Status = 26; break; }  // eta -> pi+ pi- pi0
              case 74:   { AuxTrack.Status = 27; break; }  // eta -> pi+ pi- gamma
              case 9900: { AuxTrack.Status = 28; break; }  // Cherenkov photons on particle output file
              default:   { AuxTrack.Status = 21; break; }  // all other parents/muaddi
              
            }

            float V1Diff,V2Diff,V3Diff;
            AuxTrack.Length = 0.;
            AuxTrack.T = 0.;
            AuxTrack.GenCounter = 0;
            
            // the parent relation: primary -> ... -> grandmother -> mother -> mu add. info (mu at production point) -> muon at sea
            if (ParticleID==75 || ParticleID==76 || ParticleID==85 || ParticleID==86){ // Additional muon information
              AuxMuonAddInfo=AuxTrack;
              AuxMuonAddInfo.V1 = PosX;               // x coordinate [m] (creation point)
              AuxMuonAddInfo.V2 = PosY;               // y coordinate [m] (creation point)
              AuxMuonAddInfo.V3 = PosZOrTime*0.01 + fGenPar->SiteDepth;    // z coordinate [m] (creation point)
              AuxMuonAddInfo.GenCounter = GenCounter; // hadronic generation counter of secondary
              V1Diff = AuxMuonAddInfo.V1 - OtherTracks[0].V1;
              V2Diff = AuxMuonAddInfo.V2 - OtherTracks[0].V2;
              V3Diff = AuxMuonAddInfo.V3 - OtherTracks[0].V3;
              AuxMuonAddInfo.Length = sqrt(V1Diff*V1Diff+V2Diff*V2Diff+V3Diff*V3Diff); // distance from the 1st interaction point (of the primary)
            }
            else if (ParticleID<0){ // parent particles
              if (PosZOrTime<0){ // grandmother
                AuxGrandMother=AuxTrack;
                AuxGrandMother.V1 = 0.;                   // x coordinate is not available
                AuxGrandMother.V2 = 0.;                   // y coordinate is not available
                AuxGrandMother.V3 = -PosZOrTime*0.01 + fGenPar->SiteDepth;  // z coordinate [m] is flipped
                AuxGrandMother.GenCounter = -GenCounter;  // electromagnetic generation counter of MOTHER!
                AuxTrack.GenCounter = PosXOrGenCount;     // extended electromagnetic generation counter of secondary
                AuxTrack.PenDepth = PosYOrPenDepth;       // penetration depth [g/cm^2] (penetrated matter between start and end of grandmother’s track)
              }
              else{ // mother
                AuxMother=AuxTrack;
                AuxMother.V1 = PosX;                // x coordinate [m]
                AuxMother.V2 = PosY;                // y coordinate [m]
                AuxMother.V3 = PosZOrTime*0.01 + fGenPar->SiteDepth;     // z coordinate [m]
                AuxMother.GenCounter = -GenCounter; // hadronic generation counter of mother
              }
            }
            else { // secondary particle or decaying muon at end (CORSIKA id 95/96)
              
              // because of the CURVOUT option (we consider Earth's curvature in CORSIKA)
              // we have to compute the carthesian x,y,z coordinates from the values
              // read out from the output file (see CORISKA Userguide p.95, description of CURVOUT):
              double Theta = sqrt(PosX*PosX+PosY*PosY)/RadDep;
              double SThRadDep = sin(Theta)*RadDep;
              double CTh = cos(Theta);
              double Phi = atan2(PosY,PosX);
              double SPh = sin(Phi);
              double CPh = cos(Phi);
              AuxTrack.V1 = SThRadDep*CPh;    // x coordinate
              AuxTrack.V2 = SThRadDep*SPh;    // y coordinate
              AuxTrack.V3 = fGenPar->SiteDepth+RadDep*(CTh-1.);     // z=obslev (in KM3NeT coordinates it's equal to the depth)
              AuxTrack.T = PosZOrTime;              // t [ns] since first interaction
              
              if ( pdg::IsMuon(abs(AuxTrack.PdgCode)) ) { // we calculate the track lenght if its a muon:
                V1Diff = AuxTrack.V1-AuxMuonAddInfo.V1;
                V2Diff = AuxTrack.V2-AuxMuonAddInfo.V2;
                V3Diff = AuxTrack.V3-AuxMuonAddInfo.V3;
                AuxTrack.Length = sqrt(V1Diff*V1Diff + V2Diff*V2Diff + V3Diff*V3Diff);
              }

              if ( // only the requested tracks are picked for later propagation
                  (fGenPar->SaveNu && pdg::IsNeutrino(abs(AuxTrack.PdgCode))) || // neutrino tracks
                  (fGenPar->SaveMu && pdg::IsMuon(abs(AuxTrack.PdgCode))) ||     // muon tracks
                  (fGenPar->SaveOther)                                           // tracks from other paricles
              ) {

                // setting the IDs in chronological order
                AuxGrandMother.Id = id++;
                AuxMother.Id = id++;
                if ( pdg::IsMuon(abs(AuxTrack.PdgCode)) ) AuxMuonAddInfo.Id = id++; // only increment if there is a muon
                AuxTrack.Id = id++;

                // setting the mother IDs
                AuxGrandMother.MotherId = OtherTracks[0].Id;  // corresponding primary
                AuxMother.MotherId = AuxGrandMother.Id;
                if ( pdg::IsMuon(abs(AuxTrack.PdgCode)) ) {
                  AuxMuonAddInfo.MotherId = AuxMother.Id; // mother particle is parent of the muon at production point
                  AuxTrack.MotherId = AuxMuonAddInfo.Id;
                }
                else AuxTrack.MotherId = AuxMother.Id;

                if (fGenPar->CustomRotation[0]!=0. || fGenPar->CustomRotation[1]!=0. || fGenPar->CustomRotation[2]!=0.) {
                  CustomRotation(&AuxTrack);
                  CustomRotation(&AuxMother);
                  CustomRotation(&AuxGrandMother);
                  CustomRotation(&AuxMuonAddInfo);
                }
                
                if ( AuxTrack.Status == -1 ) {  // only secondaries that will be propagated are allowed to affect the drawing area
                  // Here in fMinDistToCan we do not include the sea surface curvature, since the shower is not yet localized and we would not know what z_sea to use
                  
                  float fMinDistToCan = (fGenPar->SiteDepth-fGenPar->Can2)/fabs(AuxTrack.D3) - fGenPar->PropagationTolerance; // (shooting at the bottom edge of the can)
                  float fMaxDistToCan = (fGenPar->SiteDepth-fGenPar->Can1)/fabs(AuxTrack.D3) + fGenPar->PropagationTolerance; // (shooting at the top of the can)
                  float Range = (pow(10.,fGenPar->RangeSW.Eval(log10(AuxTrack.E)))/(fGenPar->RhoSW)) + fGenPar->MuonRangeTolerance;
                  if (Range>=fMinDistToCan) {
                    if ( fGenPar->WriteParticlesMode != 0 ) { // we save everything at sea level
                      OtherTracks.push_back(AuxGrandMother);
                      OtherTracks.push_back(AuxMother);
                      if ( pdg::IsMuon(abs(AuxTrack.PdgCode)) ) OtherTracks.push_back(AuxMuonAddInfo);
                    }
                    SecondaryTracks.push_back(AuxTrack);
                    double V2332diff = (OtherTracks[0].V2-AuxTrack.V2)*(OtherTracks[0].D3) + (AuxTrack.V3-OtherTracks[0].V3)*(OtherTracks[0].D2);
                    double V3113diff = (OtherTracks[0].V3-AuxTrack.V3)*(OtherTracks[0].D1) + (AuxTrack.V1-OtherTracks[0].V1)*(OtherTracks[0].D3);
                    double V1221diff = (OtherTracks[0].V1-AuxTrack.V1)*(OtherTracks[0].D2) + (AuxTrack.V2-OtherTracks[0].V2)*(OtherTracks[0].D1);
                    // fDista is computed as a distance of a point (muon) from the line (primary trajectory)
                    // pow(10.,...) is a fitted estimation of the lateral muon deflection (see https://git.km3net.de/simulation/gseagen/-/issues/84)
                    fDistaMax = max( fDistaMax, sqrt( V2332diff*V2332diff + V3113diff*V3113diff + V1221diff*V1221diff ) + pow(10.,(log10(fMaxDistToCan)*0.67971202-3.07154976)*log10(AuxTrack.E)+fMaxDistToCan*(-753778.571)+3.97088084) );
                  }
                  else {  // insufficient range
                    if ( fGenPar->WriteParticlesMode != 0 ) { // we don't care about parent particles for option 0
                      OtherTracks.push_back(AuxGrandMother);
                      OtherTracks.push_back(AuxMother);
                      if ( pdg::IsMuon(abs(AuxTrack.PdgCode)) ) OtherTracks.push_back(AuxMuonAddInfo);
                    }
                    OtherTracks.push_back(AuxTrack); // this muon is not worth propagating anyway
                    
                  }

                  LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"fDistaMax:"<<fDistaMax;
                }
                else {  // miscellanous secondaries (e.g. decaying muon at start/end)
                  if ( fGenPar->WriteParticlesMode != 0 ) { // we don't care about parent particles for option 0
                    OtherTracks.push_back(AuxGrandMother);
                    OtherTracks.push_back(AuxMother);
                    if ( pdg::IsMuon(abs(AuxTrack.PdgCode)) ) OtherTracks.push_back(AuxMuonAddInfo);
                  }
                  OtherTracks.push_back(AuxTrack);
                  
                }
              }
            }
            // p++;  // moves the particle_data struct pointer to the next entry (next particle)
          }
        }
        lp+=LenSubblock; // move to the next subblock
      }
    } // end of while loop (single event)
    if (FullRun || FullEvent) return vector<GSeaTrack>(); // we're done
    else {  // just in case
      LOG("GSeaCORSIKAFileFlux", pFATAL) << "Something wrong happened when trying to return tracks, exiting...";
      exit(1);
    }
  } // end of while loop
}
//___________________________________________________________________________
void GSeaCORSIKAFileFlux::ComputeWeights(GSeaEvent * SeaEvt){

  // Compute the weights
  // (accodring to https://simulation.pages.km3net.de/taglist/taglist.pdf)
  
  // fAgen is w1

  double ITheta = 2. * kPi * fabs(fGenPar->PrimCtMax - fGenPar->PrimCtMin);
  double Gamma = fabs(fGenPar->Alpha);
  double IEprim;
  
  if (Gamma == 1) IEprim = log(MaxEPrim / MinEPrim); // yes, this should be natural logarithm!
  else IEprim = (pow(MaxEPrim, 1. - Gamma) - pow(MinEPrim, 1. - Gamma)) / (1. - Gamma);

  SeaEvt->GenWeight = SeaEvt->Agen * ITheta * IEprim * pow(EPrim, Gamma) * fGenPar->TGen; // w2
  SeaEvt->EvtWeight = SeaEvt->GenWeight * CRFlux(EPrim, PrimID, fGenPar->CRModel); // w3

  this->GetEvtTime(SeaEvt->EvtTime,SeaEvt->EvtTime+1);

  LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"SeaEvt->Agen: "<<SeaEvt->Agen;
  LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"SeaEvt->GenWeight: "<<SeaEvt->GenWeight;
  LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"SeaEvt->EvtWeight: "<<SeaEvt->EvtWeight;

  return;
}
//___________________________________________________________________________
void GSeaCORSIKAFileFlux::CustomRotation(GSeaTrack *Track){
  
  float dir_x = Track->D1;
  float dir_y = Track->D2;
  float dir_z = Track->D3;
  // rotation according to https://www.sciencedirect.com/topics/engineering/direction-cosine-matrix (eq. 11.119)
  float s0 = sin(fGenPar->CustomRotation[0]);
  float s1 = sin(fGenPar->CustomRotation[1]);
  float s2 = sin(fGenPar->CustomRotation[2]);
  float c0 = cos(fGenPar->CustomRotation[0]);
  float c1 = cos(fGenPar->CustomRotation[1]);
  float c2 = cos(fGenPar->CustomRotation[2]);
  Track->D1 = (
    dir_x*c0*c1+
    dir_y*s0*c1-
    dir_z*s1
  );
  Track->D2 = (
    dir_x*(c0*s1*s2-s0*c2)+
    dir_y*(s0*s1*s2+c0*c2)+
    dir_z*c1*s2
  );
  Track->D3 = (
    dir_x*(c0*s1*c2+s0*s2)+
    dir_y*(s0*s1*c2-c0*s2)+
    dir_z*c1*c2
  );
}
//___________________________________________________________________________
int32_t GSeaCORSIKAFileFlux::CorsikaToPdg(int16_t CORSIKAId){
  std::unordered_map<int16_t, int32_t> IDS = {
        // 0 is NOT a valid PDG ID, but it is used for options not covered by
        // the PDG scheme

        //{CORSIKA, PDG}        // PARTICLE NAME
          { 1, 22 },            // photon
          { 2, -11 },           // positron
          { 3, 11 },            // electron

          { 5, -13 },           // muon +
          { 6, 13 },            // muon -
          { 7, 111 },           // pion0
          { 8, 211 },           // pion+
          { 9, -211 },          // pion-
          { 10, 130 },          // kaon0_L
          { 11, 321 },          // kaon+
          { 12, -321 },         // kaon-
          { 13, 2112 },         // neutron
          { 14, 2212 },         // proton
          { 15, -2212 },        // antiproton
          { 16, 310 },          // kaon0_S
          { 17, 221 },          // eta
          { 18, 3122 },         // Lambda
          { 19, 3222 },         // Sigma+
          { 20, 3212 },         // Sigma0
          { 21, 3112 },         // Sigma-
          { 22, 3322 },         // Xi0
          { 23, 3312 },         // Xi-
          { 24, 3334 },         // Omega-
          { 25, -2112 },        // antineutron
          { 26, -3122 },        // antiLambda
          { 27, -3222 },        // antiSigma-
          { 28, -3212 },        // antiSigma0
          { 29, -3112 },        // antiSigma+
          { 30, -3322 },        // antiXi0
          { 31, -3312 },        // antiXi+
          { 32, -3334 },        // antiOmega+

          { 48, 331 },          // eta'
          { 49, 333 },          // phi
          { 50, 223 },          // omega
          { 51, 113 },          // rho0 (assuming 770 state)
          { 52, 213 },          // rho+ (assuming 770 state)
          { 53, -213 },         // rho- (assuming 770 state)
          { 54, 2224 },         // Delta++
          { 55, 2214 },         // Delta+
          { 56, 2114 },         // Delta0
          { 57, 1114 },         // Delta-
          { 58, -2224 },        // antiDelta--
          { 59, -2214 },        // antiDelta-
          { 60, -2114 },        // antiDelta0
          { 61, -1114 },        // antiDelta+
          { 62, 313 },          // kaon*0 (assuming 892 state)
          { 63, 323 },          // kaon*+ (assuming 892 state)
          { 64, -323 },         // antikaon*-
          { 65, -313 },         // antikaon*0
          { 66, 12 },           // electron neutrino
          { 67, -12 },          // antielectron neutrino
          { 68, 14 },           // muon neutrino
          { 69, -14 },          // antimuon neutrino

          { 71, 221 },          // eta -> 2 gamma         | status=24
          { 72, 221 },          // eta -> 3 pi0           | status=25
          { 73, 221 },          // eta -> pi+ pi- pi0     | status=26
          { 74, 221 },          // eta -> pi+ pi- gamma   | status=27
          { 75, -13 },          // muon+ add. info        | status=21
          { 76, 13 },           // muon- add. info        | status=21

          { 85, -13 },          // decaying mu+ at start  | status=22
          { 86, 13 },           // decaying mu- at start  | status=22

          { 95, -13 },          // decaying mu+ at end    | status=23
          { 96, 13 },           // decaying mu- at end    | status=23

          { 116, 421 },         // D0
          { 117, 411 },         // D+
          { 118, -411 },        // antiD-
          { 119, -421 },        // antiD0
          { 120, 431 },         // D_s+
          { 121, -431 },        // antiD_s-
          { 122, 441 },         // eta_c (assuming always eta_c (1S))
          { 123, 423 },         // D*0 (assuming 2007 state)
          { 124, 413 },         // D*+ (assuming 2010 state)
          { 125, -413 },        // antiD*-
          { 126, -423 },        // antiD*0
          { 127, 433 },         // D_s*-
          { 128, -433 },        // antiD_s*-

          { 130, 443 },         // J/psi
          { 131, -15 },         // tau+
          { 132, 15 },          // tau-
          { 133, 16 },          // nutau
          { 134, -16 },         // antinutau

          { 137, 4122 },        // Lambda_c+
          { 138, 4232 },        // Xi_c+
          { 139, 4132 },        // Xi_c0
          { 140, 4222 },        // Sigma_c++
          { 141, 4212 },        // Sigma_c+
          { 142, 4112 },        // Sigma_c0
          { 143, 4322 },        // Xi_c'+
          { 144, 4312 },        // Xi_c'0
          { 145, 4332 },        // Omega_c0

          { 149, -4122 },       // antiLambda_c-
          { 150, -4232 },       // antiXi_c-
          { 151, -4132 },       // antiXi_c0
          { 152, -4222 },       // antiSigma_c--
          { 153, -4212 },       // antiSigma_c-
          { 154, -4112 },       // antiSigma_c0
          { 155, -4322 },       // antiXi_c'-
          { 156, -4312 },       // antiXi_c'0
          { 157, -4332 },       // antiOmega_c0

          { 161, 4224 },        // Sigma_c*++
          { 162, 4214 },        // Sigma_c*+
          { 163, 4114 },        // Sigma_c*0

          { 171, -4224 },       // antiSigma_c*--
          { 172, -4214 },       // antiSigma_c*-
          { 173, -4114 },       // antiSigma_c*0

          { 176, 511 },         // B0
          { 177, 521 },         // B+
          { 178, -521 },        // antiB-
          { 179, -511 },        // antiB0
          { 180, 531 },         // B_s0
          { 181, -531 },        // antiB_s0
          { 182, 541 },         // B_c+
          { 183, -541 },        // antiB_c-
          { 184, 5122 },        // Lambda_b0
          { 185, 5112 },        // Sigma_b-
          { 186, 5222 },        // Sigma_b+
          { 187, 5232 },        // Xi_b0
          { 188, 5132 },        // Xi_b-
          { 189, 5332 },        // Omega_b-
          { 190, -5122 },       // antiLambda_b0
          { 191, -5112 },       // antiSigma_b+
          { 192, -5222 },       // antiSigma_b-
          { 193, -5232 },       // antiXi_b0
          { 194, -5132 },       // antiXi_b+
          { 195, -5332 },       // antiOmega_b+

          // Nuclei examples:    A x 100 + Z
          { 101, 1000010010 },  // 'hydrogen'
          { 201, 1000010020 },  // 'deuteron'},
          { 301, 1000010030 },  // 'tritium',
          { 302, 1000020030 },  // 'helium3',
          { 402, 1000020040 },  // 'alpha',
          { 703, 1000030070 },  // 'lithium',
          { 904, 1000040090 },  // 'beryllium',
          { 1105, 1000050110 }, // 'boron',
          { 1206, 1000060120 }, // 'carbon',
          { 1407, 1000070140 }, // 'nitrogen',
          { 1608, 1000080160 }, // 'oxygen',
          { 2713, 1000130270 }, // 'aluminium',
          { 2814, 1000140280 }, // 'silicon',
          { 3216, 1000160320 }, // 'sulfur',
          { 4020, 1000200400 }, // 'calcium',
          { 5626, 1000260560 }, // 'iron',
          { 5828, 1000280580 }, // 'nickel',

          { 9900, 1 },    // Cherenkov photons on particle output file
                          // (NOT there without the CERENKOV option, which )
                          // (is not used for KM3NeT CORSIKA simulations)
                          // | status=-28
  };

  if ((CORSIKAId < 200) || (CORSIKAId == 9900)) return IDS[CORSIKAId]; // covered above
  else {                      // nuclei
    uint8_t nZ = CORSIKAId%100;
    uint8_t nA = (CORSIKAId-nZ)/100;
    char out_id[10];
    sprintf(out_id, "100%03d%03d0", nZ, nA);
    int32_t out_id_int = std::stoi (out_id,nullptr,10);
    return out_id_int;
  }
}
//___________________________________________________________________________
double GSeaCORSIKAFileFlux::CorsikaRestMass(int16_t CORSIKAId){
  std::unordered_map<int16_t, double> Masses = {
        // All masses are taken from PDG booklet 2020 (https://pdg.lbl.gov)
        // unless noted otherwise
        // neutrino masses are neglected

        //{CORSIKA, rest mass [GeV/c^2]}    // PARTICLE NAME
          { 1, 0. },                        // photon
          { 2, 0.0005109989461 },           // positron
          { 3, 0.0005109989461 },           // electron

          { 5, 0.1056583745 },              // muon +
          { 6, 0.1056583745 },              // muon -
          { 7, 0.1349768 },                 // pion0
          { 8, 0.13957039 },                // pion+
          { 9, 0.13957039 },                // pion-
          { 10, 0.497611 },                 // kaon0_L
          { 11, 0.493677 },                 // kaon+
          { 12, 0.493677 },                 // kaon-
          { 13, 0.939565413 },              // neutron
          { 14, 0.938272081 },              // proton
          { 15, 0.938272081 },              // antiproton
          { 16, 0.497611 },                 // kaon0_S
          { 17, 0.547862 },                 // eta
          { 18, 1.115683 },                 // Lambda
          { 19, 1.18937 },                  // Sigma+
          { 20, 1.192642 },                 // Sigma0
          { 21, 1.197449 },                 // Sigma-
          { 22, 1.31486 },                  // Xi0
          { 23, 1.32171 },                  // Xi-
          { 24, 1.67245 },                  // Omega-
          { 25, 0.939565413 },              // antineutron
          { 26, 1.115683 },                 // antiLambda
          { 27, 1.18937 },                  // antiSigma-
          { 28, 1.192642 },                 // antiSigma0
          { 29, 1.197449 },                 // antiSigma+
          { 30, 1.31486 },                  // antiXi0
          { 31, 1.32171 },                  // antiXi+
          { 32, 1.67245 },                  // antiOmega+

          { 48, 0.95778 },                  // eta' (assuming 958 state)
          { 49, 1.019461 },                 // phi (assuming 1020 state)
          { 50, 0.78265 },                  // omega (assuming 782 state)
          { 51, 0.77526 },                  // rho0 (assuming 770 state)
          { 52, 0.77526 },                  // rho+ (assuming 770 state)
          { 53, 0.77526 },                  // rho- (assuming 770 state)
          { 54, 2224 },                     // Delta++
          { 55, 1.210 },                    // Delta+ (1232 state)
          { 56, 2114 },                     // Delta0
          { 57, 1.610 },                    // Delta- (1620 state)
          { 58, -2224 },                    // antiDelta--
          { 59, 1.210 },                    // antiDelta-
          { 60, -2114 },                    // antiDelta0
          { 61, 1.610 },                    // antiDelta+
          { 62, 0.89555 },                  // kaon*0 (assuming 892 state and hadroproduction)
          { 63, 0.89166 },                  // kaon*+ (assuming 892 state and hadroproduction)
          { 64, 0.89166 },                  // antikaon*-
          { 65, 0.89555 },                  // antikaon*0
          { 66, 0. },                       // electron neutrino
          { 67, 0. },                       // antielectron neutrino
          { 68, 0. },                       // muon neutrino
          { 69, 0. },                       // antimuon neutrino

          { 71, 0.547862 },                 // eta -> 2 gamma
          { 72, 0.547862 },                 // eta -> 3 pi0
          { 73, 0.547862 },                 // eta -> pi+ pi- pi0
          { 74, 0.547862 },                 // eta -> pi+ pi- gamma
          { 75, 0.1056583745 },             // muon+ add. info
          { 76, 0.1056583745 },             // muon- add. info

          { 85, 0.1056583745 },             // decaying mu+ at start
          { 86, 0.1056583745 },             // decaying mu- at start

          { 95, 0.1056583745 },             // decaying mu+ at end
          { 96, 0.1056583745 },             // decaying mu- at end

          { 116, 1.86483 },                 // D0
          { 117, 1.86965 },                 // D+
          { 118, 1.86965 },                 // antiD-
          { 119, 1.86483 },                 // antiD0
          { 120, 1.96834 },                 // D_s+
          { 121, 1.96834 },                 // antiD_s-
          { 122, 2.9839 },                  // eta_c (assuming always eta_c (1S))
          { 123, 2.00685 },                 // D*0 (assuming 2007 state)
          { 124, 2.01026 },                 // D*+ (assuming 2010 state)
          { 125, 2.01026 },                 // antiD*-
          { 126, 2.00685 },                 // antiD*0
          { 127, 2.1122 },                  // D_s*-
          { 128, 2.1122 },                  // antiD_s*+

          { 130, 3.096900 },                // J/psi
          { 131, 1.77686 },                 // tau+
          { 132, 1.77686 },                 // tau-
          { 133, 0. },                      // nutau
          { 134, 0. },                      // antinutau

          { 137, 2.28646 },                 // Lambda_c+
          { 138, 2.46794 },                 // Xi_c+
          { 139, 2.47090 },                 // Xi_c0
          { 140, 2.45397 },                 // Sigma_c++
          { 141, 2.4529 },                  // Sigma_c+ (2455 state)
          { 142, 2.45375 },                 // Sigma_c0
          { 143, 2.5784 },                  // Xi_c'+
          { 144, 2.5792 },                  // Xi_c'0
          { 145, 2.6952 },                  // Omega_c0

          { 149, 2.28646 },                 // antiLambda_c-
          { 150, 2.46794 },                 // antiXi_c-
          { 151, 2.47090 },                 // antiXi_c0
          { 152, 2.45397 },                 // antiSigma_c--
          { 153, 2.4529 },                  // antiSigma_c-
          { 154, 2.45375 },                 // antiSigma_c0
          { 155, 2.5784 },                  // antiXi_c'-
          { 156, 2.5792 },                  // antiXi_c'0
          { 157, 2.6952 },                  // antiOmega_c0

          { 161, 2.5179 },                  // Sigma_c*++ from https://en.wikipedia.org/wiki/List_of_baryons
          { 162, 2.5175 },                  // Sigma_c*+  from https://en.wikipedia.org/wiki/List_of_baryons
          { 163, 2.5188 },                  // Sigma_c*0  from https://en.wikipedia.org/wiki/List_of_baryons

          { 171, 2.5179 },                  // antiSigma_c*--  from https://en.wikipedia.org/wiki/List_of_baryons
          { 172, 2.5175 },                  // antiSigma_c*-  from https://en.wikipedia.org/wiki/List_of_baryons
          { 173, 2.5188 },                  // antiSigma_c*0  from https://en.wikipedia.org/wiki/List_of_baryons

          { 176, 5.27965 },                 // B0
          { 177, 5.27934 },                 // B+
          { 178, 5.27934 },                 // antiB-
          { 179, 5.27965 },                 // antiB0
          { 180, 5.36688 },                 // B_s0
          { 181, 5.36688 },                 // antiB_s0
          { 182, 6.2749 },                  // B_c+
          { 183, 6.2749 },                  // antiB_c-
          { 184, 5.61960 },                 // Lambda_b0
          { 185, 5.81564 },                 // Sigma_b-
          { 186, 5.81056 },                 // Sigma_b+
          { 187, 5.7919 },                  // Xi_b0
          { 188, 5.7970 },                  // Xi_b-
          { 189, 6.0461 },                  // Omega_b-
          { 190, 5.61960 },                 // antiLambda_b0
          { 191, 5.81564 },                 // antiSigma_b+
          { 192, 5.81056 },                 // antiSigma_b-
          { 193, 5.7919 },                  // antiXi_b0
          { 194, 5.7970 },                  // antiXi_b+
          { 195, 6.0461 },                  // antiOmega_b+

          { 9900, 0. },     // Cherenkov photons on particle output file
                            // (NOT there without the CERENKOV option, which )
                            // (is not used for KM3NeT CORSIKA simulations)
  };

  double Mass,EBinding,CorrMass;
  if ((CORSIKAId < 200) || (CORSIKAId == 9900)) CorrMass = Masses[CORSIKAId];      // covered above
  else {                      // nuclei
    uint8_t nZ = CORSIKAId%100;       // number of protons
    uint8_t nA = (CORSIKAId-nZ)/100;  // number of protons + neutrons
    uint8_t nN = nA - nZ;             // number of neutrons
    // without binding energy
    Mass = Masses[13] * nN + Masses[14] * nZ;

    // binding energy correction calculated from the fit to the
    // Bethe–Weizsäcker mass formula:
    // Wapstra, A. H. (1958), 'Atomic Masses of Nuclides', In Flügge, S. (ed.),
    // External Properties of Atomic Nuclei, Encyclopedia of Physics, 8/38/1,
    // Springer, pp. 1–37, Bibcode:1958HDP....38....1W,
    // doi:10.1007/978-3-642-45901-6_1, ISBN 978-3-642-45902-3
    //
    // also described in https://en.wikipedia.org/wiki/Semi-empirical_mass_formula
    double cbrtnZ = cbrt(nZ);
    EBinding = 14.1 * nZ - 0.595 * nZ*nZ * (1./cbrtnZ) - 13. * cbrtnZ*cbrtnZ - 19. * (nZ-nN)*(nZ-nN) / nZ;

    // Here we add the pairing term:
    if ( nZ == 0 && nN > 0 ) EBinding = 0.;                                                     // multi-neutron clusters
    else if ( nA % 2 == 1 ) EBinding = EBinding;                                                // nA odd
    else if ( nZ % 2 == 0 && nN % 2 == 0 ) EBinding = EBinding + 33.5 * pow(nZ,-0.75); // nA even
    else if ( nZ % 2 == 1 && nN % 2 == 1 ) EBinding = EBinding - 33.5 * pow(nZ,-0.75); // nA even
    else {
      LOG("GSeaCORSIKAFileFlux", pFATAL) << "Unknown nucleus (A,Z): (" << nA << "," << nZ << "), exiting...";
      exit(1);
    }

    if ( EBinding > 0. ) EBinding=1.e-3*EBinding; // convert from MeV to GeV
    else EBinding=0.;

    CorrMass = Mass-EBinding;

  }
  return CorrMass;
}
//___________________________________________________________________________
double GSeaCORSIKAFileFlux::CRFlux(double PrimEnergy, int16_t PrimCORSIKAId, string CRFluxModel){
  // implementation adapted from https://crfluxmodels.readthedocs.io/en/latest/_modules/models.html#PrimaryFlux
  // output in m^2 / {s * sr * GeV}

  double Flux = 0.0;

  if (CRFluxModel=="GST3") {

    // T. K. Gaisser, T. Stanev, and S. Tilav, arXiv:1303.3565, (2013)

    const double RigCut[3] = {120e3,4e6,1.3e9}; // Rigidity cutoffs

    std::unordered_map<uint16_t, uint8_t> NuclToID = {
      { 14, 0 },    // p
      { 402, 1 },   // He
      { 1206, 2 },  // C
      { 1608, 3 },  // O
      { 5626, 4 },  // Fe
    };

    double Params[5][3][3] = {
      { // p (14)
        {7000., 1.66, 1.},
        {150., 1.4, 1.},
        {14., 1.4, 1.}
      },{ // He (402)
        {3200., 1.58, 2.},
        {65., 1.3, 2.},
        {0., 1.4, 2.}
      },{ // C (1206)
        {100., 1.4, 6.},
        {6., 1.3, 6.},
        {0., 1.4, 6.}
      },{ // O (1608)
        {130., 1.4, 8.},
        {7., 1.3, 8.},
        {0., 1.3, 8.}
      },{ // Fe (5626)
        {60., 1.3, 26.},
        {2.3, 1.2, 26.},
        {0.025, 1.2, 26.}
      }
};

    for (unsigned int i = 0; i < 3; i++) {
      Flux += Params[NuclToID[PrimCORSIKAId]][i][0] * pow(PrimEnergy,(-Params[NuclToID[PrimCORSIKAId]][i][1] - 1.0)) * exp(-PrimEnergy / (Params[NuclToID[PrimCORSIKAId]][i][2] * RigCut[i]));
    }

  }
  else {
    LOG("GSeaCORSIKAFileFlux", pFATAL) << "Undefined primary CR flux model, exiting...";
    exit(1);
  }

  return Flux;
}
