//____________________________________________________________________________
/*!

\program gCalcCross

\brief   Calculate cross section from pre-computed GENIE splines

\author  Carla Distefano
         INFN - LNS

\created June 13, 2017

\cpright  Copyright (c) 2017-2019, The KM3NeT Collaboration


*/
//____________________________________________________________________________


#include <cassert>
#include <cstdlib>
#include <cctype>
#include <string>
#include <vector>
#include <sstream>
#include <map>
#include <cstdio>
#include <iostream>
#include <fstream>
#include <cmath>
#include <ctime>
#include <algorithm>

#include "libxml/parser.h"
#include "libxml/xmlmemory.h"
#include "libxml/xmlreader.h"

#include <TSystem.h>
#include <TRandom.h>
#include <TGraph.h>
#include <TFile.h>


// GENIE

#ifdef _GENIEVERSIONGTEQ3__
#include "Framework/EventGen/GEVGDriver.h"
#include "Framework/Conventions/Units.h"
#include "Framework/GHEP/GHepRecord.h"
#include "Framework/Messenger/Messenger.h"
#include "Framework/Numerical/Spline.h"
#include "Framework/ParticleData/PDGLibrary.h"
#include "Framework/ParticleData/PDGUtils.h"
#include "Framework/Utils/AppInit.h"
#include "Framework/Utils/CmdLnArgParser.h"
#include "Framework/Utils/RunOpt.h"
#include "Framework/Utils/StringUtils.h"
#else
#include "EVGDrivers/GEVGDriver.h"
#include "Conventions/Units.h"
#include "GHEP/GHepRecord.h"
#include "Messenger/Messenger.h"
#include "Numerical/Spline.h"
#include "PDG/PDGLibrary.h"
#include "PDG/PDGUtils.h"
#include "Utils/AppInit.h"
#include "Utils/CmdLnArgParser.h"
#include "Utils/RunOpt.h"
#include "Utils/StringUtils.h"
#endif



using namespace std;
using namespace genie;

void GetCommandLineArgs (int argc, char ** argv);
void PrintSyntax        (void);
void ReadTargetFile     (void);


string InpXSecFile="$GSEAGEN/dat/gxspl-seawater-muon.xml";

void Initialize();

double gEvMin=1.;
double gEvMax=5000.;
int gNuType=14;
int gNP=100;
string gInter="Default";
string gTarget;

map<int, double> MediaComp;

string gTargetName="Water";
double gTargetRho=1.;

string gOutPut="xsec.root";

string gTune="G18_02a_00_000";

//___________________________________________________________________
int main(int argc, char ** argv)
{

  MediaComp.insert(map<int, double>::value_type(1000080160,0.8888));            // O16
  MediaComp.insert(map<int, double>::value_type(1000010010,0.1112));            // H1

  GetCommandLineArgs (argc, argv);

  Initialize();

  int TgtPdgCode;
  double TgtPercentage;

  GEVGDriver evg_driver;
  InitialState init_state;
  const Spline * splsum;

  evg_driver.SetEventGeneratorList(gInter);

  size_t nComp=MediaComp.size();
  TGraph * gr[nComp];
  double PA[nComp];
  int A[nComp];

  double x,y,ytot;

  unsigned iComp=0;

  map<int,double>::iterator iter = MediaComp.begin();
  for ( ; iter != MediaComp.end(); ++iter) {

    TgtPdgCode=iter->first;
    TgtPercentage=iter->second;
    A[iComp]=pdg::IonPdgCodeToA(TgtPdgCode);
    PA[iComp]=TgtPercentage/(double)pdg::IonPdgCodeToA(TgtPdgCode);
    init_state.SetPdgs(TgtPdgCode, gNuType);
    evg_driver.Configure(init_state);

//  Check if the user defined Emin and Emax are outside the valid energy range
    Range1D_t rangeE = evg_driver.ValidEnergyRange();
    if(gEvMin<rangeE.min){
      LOG("gCalcCross", pWARN) << "Emin lower than valid Emin value "<<rangeE.min<<" GeV, redefined.";
      gEvMin=rangeE.min;
    }
    if(gEvMax>rangeE.max){
      LOG("gCalcCross", pWARN) << "Emax greater than valid Emax value "<<rangeE.max<<" GeV, redefined.";
      gEvMax=rangeE.max;
    }


    evg_driver.CreateSplines();
    evg_driver.CreateXSecSumSpline (gNP, gEvMin, gEvMax);
    splsum = evg_driver.XSecSumSpline();

    TParticlePDG * p = PDGLibrary::Instance()->Find(TgtPdgCode);

    gr[iComp] = splsum->GetAsTGraph(gNP,true,true,1.,1./units::m2);
    gr[iComp]->SetTitle(p->GetName());

    iComp++;
  }

  TFile f(gOutPut.c_str(),"RECREATE");

  TGraph XSecTot;


  string XSecTitle="Cross Section in ";
  XSecTitle.append(gTargetName);
  XSecTitle.append(" ;Ev (GeV); #sigma (m^{2})");
  XSecTot.SetTitle(XSecTitle.c_str());

  string IntLenTitle="Interaction Length in ";
  IntLenTitle.append(gTargetName);
  IntLenTitle.append(" (m);Ev (GeV); L (m)");

  TGraph IntLen;
  IntLen.SetTitle(IntLenTitle.c_str());


  for (int i=0; i<gNP; i++){
    ytot=0.;
    for(iComp=0; iComp<nComp; iComp++){
      gr[iComp]->GetPoint(i,x,y);
      ytot+=PA[iComp]*y*x;
    }
    if(ytot>0){
      XSecTot.SetPoint(i,x,ytot);
      IntLen.SetPoint(i,x,1E-6/(gTargetRho*6.02214179E+23*ytot)); // gTargetRho in g/cm^3
    }
  }

  if(nComp>1){
    for(iComp=0; iComp<nComp; iComp++){
      for (int i=0; i<gNP; i++){
        gr[iComp]->GetPoint(i,x,y);
        if(y<=0)gr[iComp]->RemovePoint(i);
        else gr[iComp]->SetPoint(i,x,y*x/A[iComp]);
      }

      string nucleus=gr[iComp]->GetTitle();
      XSecTitle="Cross Section in ";
      XSecTitle.append(nucleus.c_str());
      XSecTitle.append(" ;Ev (GeV); #sigma (m^{2})");
      gr[iComp]->SetTitle(XSecTitle.c_str());

      XSecTitle="xsec-";
      XSecTitle.append(nucleus.c_str());

      gr[iComp]->Write(XSecTitle.c_str());
    }
  }

  XSecTot.Write("xsec");
  IntLen.Write("lint");
  f.Close();

  cout<<endl;


  return 0;
}
//___________________________________________________________________
void GetCommandLineArgs(int argc, char ** argv)
{


  LOG("gCalcCross", pINFO)
     << "*** Parsing command line arguments";

  LOG("gCalcCross", pINFO) << "Parsing command line arguments";

  // Common run options. Set defaults and read.
  RunOpt::Instance()->ReadFromCommandLine(argc,argv);

  // Parse run options for this app

  CmdLnArgParser parser(argc,argv);

  if(parser.OptionExists('e')){         // neutrino energy range

    LOG("gCalcCross", pINFO) << "Reading neutrino energy range";
    string nue =  parser.ArgAsString('e');
    // must be a comma separated set of values
    if(nue.find(",") != string::npos) {
      // split the comma separated list
      vector<string> nurange = utils::str::Split(nue, ",");
      assert(nurange.size() == 2);
      double emin = atof(nurange[0].c_str());
      double emax = atof(nurange[1].c_str());
      if(emin==emax)emax+=0.001;
      assert(emax>emin && emin>=0.);
      gEvMin = emin;
      gEvMax = emax;
    }
  }
  if(parser.OptionExists('n')){ // nu type
    LOG("gCalcCross", pINFO) << "Reading nu type";
    gNuType = parser.ArgAsInt('n');
  }
  if(parser.OptionExists('b')){ // npoints
    LOG("gCalcCross", pINFO) << "Reading number of energy bins";
    gNP = parser.ArgAsInt('b');
  }
  if(parser.OptionExists('i')){ // interaction type
    LOG("gCalcCross", pINFO) << "Reading interaction points";
    gInter = parser.ArgAsString('i');
  }
  if(parser.OptionExists('o')){ // output file name
    LOG("gCalcCross", pINFO) << "Reading the output file name";
    gOutPut = parser.ArgAsString('o');
  }
  if(parser.OptionExists('t')){ // target
    LOG("gCalcCross", pINFO) << "Reading the target file";
    gTarget = parser.ArgAsString('t');
    ReadTargetFile();
  }
  if(parser.OptionExists('f')){ // target
    LOG("gCalcCross", pINFO) << "Reading the cross section file";
    InpXSecFile = parser.ArgAsString('f');
  }

#ifdef _GENIEVERSIONGTEQ3__
  if(parser.OptionExists('m')){ // tune
    LOG("gCalcCross", pINFO) << "Reading the genie tune";
    gTune = parser.ArgAsString('m');
  }
#endif


  if(parser.OptionExists('h')){ // print help
    PrintSyntax();
    exit(0);
  }


   LOG("gCalcCross", pNOTICE)
     << "\n"
     << "\n ****** Compute cross section in water ****** "
     << "\n\t Using energy range = (" << gEvMin << " GeV, " << gEvMax << " GeV)"
     << "\n\t Using neutrino type = " << gNuType
     << "\n\t Using interaction type = " << gInter
     << "\n\t Using number of energy bins = " << gNP
     << "\n\n";

}
//_________________________________________________________________________________
void ReadTargetFile(void){


  MediaComp.clear();

    xmlTextReaderPtr reader;

    string filename=gTarget;

    if(!(access(filename.c_str(), F_OK) != -1)){
      LOG("gCalCross", pFATAL) << filename <<" does not exist; exiting...";
      exit(1);
    }
    else
      LOG("gSeaNuEvGen", pINFO) <<"Reading the target file "<< filename;

    reader = xmlNewTextReaderFilename(filename.c_str());

    string smedia,selem,sdensity;
    int pdg=0;

    while(xmlTextReaderRead(reader)==1){

      xmlChar * name  = xmlTextReaderName     (reader);
      xmlChar * value = xmlTextReaderValue    (reader);
      int       type  = xmlTextReaderNodeType (reader);
      int       depth = xmlTextReaderDepth    (reader);

      if(depth==0 && type==1){
        if(xmlStrcmp(name, (const xmlChar *) "media_comp")){
        LOG("gSeaNuEvGen", pFATAL) << " invalid root element in filename " << filename <<"; exiting... ";
        exit(1);
       }
      }
      if(!xmlStrcmp(name, (const xmlChar*) "param_set" ) && type==1){
        xmlChar * xmedia   = xmlTextReaderGetAttribute(reader,(const xmlChar*)"media");
        smedia = utils::str::TrimSpaces((const char *)xmedia);

        xmlChar * xdensity = xmlTextReaderGetAttribute(reader,(const xmlChar*)"density");
        sdensity = utils::str::TrimSpaces((const char*)xdensity);

        gTargetName=smedia;
        gTargetRho=(double)atof(sdensity.c_str());

        xmlFree(xmedia);
      }

      if(!xmlStrcmp(name, (const xmlChar *) "param" ) && type==1) {
        xmlChar * xelem   = xmlTextReaderGetAttribute(reader,(const xmlChar*)"name");
        selem = utils::str::TrimSpaces((const char *)xelem);
        xmlFree(xelem);
      }

      if(depth==3){

        pdg=atoi(selem.c_str());
        MediaComp.insert(map<int, double>::value_type(pdg,atof((const char *)value)));
      }
      xmlFree(name);
      xmlFree(value);
    }


}
//_________________________________________________________________________________
void PrintSyntax(void)
{
  LOG("gCalcCross", pFATAL)
     << "\n\n"
     << "gCalcCross                     \n"
     << "    [-e emin,emax]             \n"
     << "    [-n nutype]                \n"
     << "    [-b nbins]                 \n"
     << "    [-f xsec.xml]              \n"
     << "    [-i interaction]           \n"
#ifdef _GENIEVERSIONGTEQ3__
     << "    [-m genie_tune]            \n"
#endif
     << "    [-t target.xml]            \n"
     << "    [-o output.root]           \n"
     << "    [-h]                       \n"
     << "\n"
     << "\n **Code Options**"
     << "\n"
     << "\n [] Denotes an optional argument."
     << "\n"
     << "\n -e        Specifies the neutrino energy in GeV [default: 1,5000]. "
     << "\n           Must be a comma-separated pair of numbers, eg. `-e 100,500' "
     << "\n -n        Specifies the neutrino type in the PDG code (12: nu_e, -12: anti_nu_e, 14: nu_mu,"
     << "\n           -14: anti_nu_mu, 16: nu_tau, -16: anti_nu_tau) [default: 14]. "
     << "\n"
     << "\n -b        Specifies the numebr of energy bins for xsec calculation [default: 100]. "
     << "\n"
     << "\n -f        Specifies the input XML cross section spline file. [default: '$GSEAGEN/dat/gxspl-seawater-muon.xml']"
     << "\n"
     << "\n -i        Interaction type. Valid settings are the XML block names appearing in"
     << "\n           config/EventGeneratorListAssembler.xml [default: 'Default']."
     << "\n"
     << "\n           eg. '-i  Default' (the default considers all interaction types)"
     << "\n           eg. '-i  CC' "
     << "\n           eg. '-i  NC' "
     << "\n"
#ifdef _GENIEVERSIONGTEQ3__
     << "\n -m        Specifies the genie tune [default: G18_02a_00_000]. "
     << "\n"
     << "\n           eg. '-t G18_02a_00_000'"
     << "\n"
#endif
     << "\n -t        Specifies file name containing the target mixture [default: water]. "
     << "\n"
     << "\n           eg. '-t $GSEAGEN/src/tools/gCalcCross/water.xml'"
     << "\n"
     << "\n -o        Specifies the output file name containing the cross sections [default: xsec.root]. "
     << "\n"
     << "\n -h        Prints this help"
     << "\n";


}
//_________________________________________________________________________________
void Initialize()
{

#ifdef _GENIEVERSIONGTEQ3__
  RunOpt::Instance()->SetEventGeneratorList(gInter);
  RunOpt::Instance()->SetTuneName(gTune);
  if ( ! RunOpt::Instance()->Tune() ) {
    LOG("GGenerateEventGenie", pFATAL) << " No TuneId in RunOption";
    exit(-1);
  }
  RunOpt::Instance()->BuildTune();
#endif

  // Initialization of random number generators, cross-section table,
  // messenger thresholds, cache file
  utils::app_init::MesgThresholds(RunOpt::Instance()->MesgThresholdFiles());
  utils::app_init::CacheFile(RunOpt::Instance()->CacheFile());
  utils::app_init::RandGen((long int)1);
  utils::app_init::XSecTable(InpXSecFile, false);

  // Set GHEP print level
  GHepRecord::SetPrintLevel(RunOpt::Instance()->EventRecordPrintLevel());
  return;
}

