//____________________________________________________________________________
/*!

\program gFileConvert

\brief   Convert the native gSeaGen output file in other formats

\author  Carla Distefano
         INFN - LNS

\created March 24, 2022



*/
//____________________________________________________________________________


#include <cassert>
#include <cstdlib>
#include <cctype>
#include <string>
#include <vector>
#include <sstream>
#include <map>
#include <cstdio>
#include <iostream>
#include <fstream>
#include <cmath>
#include <ctime>
#include <algorithm>

#include "libxml/parser.h"
#include "libxml/xmlmemory.h"
#include "libxml/xmlreader.h"

#include <TSystem.h>
#include <TRandom.h>
#include <TGraph.h>
#include <TFile.h>

#ifdef _ANTARES_ENABLED__
#include "OutputWriters/EvtWrite.h"
#endif
#ifdef _KM3NET_ENABLED__
#include "OutputWriters/KM3NeTWrite.h"
#endif
#include "OutputWriters/SeaNtpWriter.h"

#include "SeaEvent/GBinParam.h"

using namespace std;

void GetCommandLineArgs (int argc, char ** argv);
void PrintSyntax        (void);


string gFileName;
string gOutFormat;

bool             gOptPDG       = false;

//___________________________________________________________________
int main(int argc, char ** argv)
{

  GetCommandLineArgs (argc, argv);

  TTree * tHeader = 0;
  TTree * tEvents = 0;

  GenParam * gGenPar = 0;
  GSeaEvent * gSeaEvent = 0;

  TFile * inpfile = new TFile(gFileName.c_str(),"READ");

  tHeader = dynamic_cast <TTree *> (inpfile->Get("Header"));
  tEvents = dynamic_cast <TTree *> (inpfile->Get("Events"));

  tHeader->SetBranchAddress("Header", &gGenPar);
  tEvents->SetBranchAddress("Events", &gSeaEvent);
  tHeader->GetEntry(0);



#ifdef _ANTARES_ENABLED__
  EvtWrite * evtwrt = 0;
#endif

#ifdef _KM3NET_ENABLED__
  KM3NeTWrite * km3net_writer = 0;
#endif

  SeaNtpWriter * ntprt = 0;

  if(gOutFormat=="st"){
    ntprt = new SeaNtpWriter(gGenPar, "SimpleTree");
  }
#ifdef _KM3NET_ENABLED__
  else if(gOutFormat=="km3net"){
    km3net_writer = new KM3NeTWrite( gGenPar );
    km3net_writer->WriteHeader();
  }
#endif
#ifdef _ANTARES_ENABLED__
  else if(gOutFormat=="antares"){
    evtwrt = new EvtWrite(gGenPar,gOptPDG);
    evtwrt->WriteHeader();
  }  
#endif
  else {
    cout<<"Specified output format not known, exiting..."<<endl;
    exit(1);
  }



  for(Int_t i=0; i<tEvents->GetEntries(); i++){
    tEvents->GetEntry(i); 
    if(gOutFormat=="st"){
      ntprt->WriteEvent(gSeaEvent);
    }
#ifdef _KM3NET_ENABLED__
    else if(gOutFormat=="km3net")km3net_writer->WriteEvent(gSeaEvent);
#endif
#ifdef _ANTARES_ENABLED__
    else if(gOutFormat=="antares")evtwrt->WriteEvent(gSeaEvent);
#endif
  }


  if(gOutFormat=="st") delete ntprt;
#ifdef _KM3NET_ENABLED__
  else if(gOutFormat=="km3net") delete km3net_writer;
#endif
#ifdef _ANTARES_ENABLED__
  else if(gOutFormat=="antares") delete evtwrt;
#endif


  return 0;
}
//___________________________________________________________________
void GetCommandLineArgs(int argc, char ** argv)
{

  size_t apos;
  string opt,value;


  cout<<"*** Parsing command line arguments"<<endl;
  for (int i = 1; i < argc; i++)
  {
    string arg(argv[i]);
    if (arg.compare(0,1,"-")==0){
      apos=arg.find(" ");
      opt=arg.substr(0,apos);

      if(opt.compare("-i")==0){         // input file name
        i++;
	gFileName = argv[i];
      }
      if(opt.compare("-f")==0){         // output format
        i++;
        gOutFormat = argv[i];
      }
      if(opt.compare("-h")==0){         // print help
        PrintSyntax();
	exit(1);
      }
      
    }
  }

  if(argc==1){
    PrintSyntax();
    exit(1);
  }

  if(gFileName.size()==0){
    cout<<"No input file name specified..., exiting"<<endl;
    PrintSyntax();
    exit(1);
  }
  
  
  if(gOutFormat.size()==0){
    cout<<"No output format specified..., exiting"<<endl;
    PrintSyntax();
    exit(1);
  }


}

//_________________________________________________________________________________
void PrintSyntax(void)
{

  cout
   << "\n"
   << "\n **Syntax**"
   << "\n gFileConvert "
   << "\n           -i input_file_name"
   << "\n           -f output_file_format"
   << "\n"

   << "\n **Code Options**"
   << "\n"
   << "\n -i        Specifies the input file name input_file_name "// [default: km3net].
   << "\n           The input file must be in the gSeaGen EventTree format "
   << "\n"
   << "\n -f        Specifies the output file format output_file_format "// [default: km3net].
   << "\n           Possible choices are:"
   << "\n           - st           : SimpleTree format"

#ifdef _ANTARES_ENABLED__
   << "\n           - antares      : ANTARES evt format"
#endif

#ifdef _KM3NET_ENABLED__
   << "\n           - km3net       : KM3NeT ROOT format"
#endif

   << "\n\n";
}
